// Zustzliche Standard-Header drfen hier eingebunden werden.
#include <cstring>
#include <list>

namespace Aremion
{
    double computeScore(const PlanetMap& planetMap,
                        const Border& border)
    {
        // zunchst die Lsung auf Gltigkeit prfen
        for(int x = 0; x < MAP_WIDTH; ++x)
        {
            if(border[x] <= 0 || border[x] >= MAP_HEIGHT)
            {
                char msg[256];
                sprintf(msg, "Ungltiger Grenzverlauf! border[%d] = %d", x, border[x]);
                throw std::runtime_error(msg);
            }
        }

        // Die Bewertung ergibt sich aus zwei Teilen:
        // 1.) "Lnge" der Grenze (s.u.)
        // 2.) Qualitt der Aufteilung (s.u.)

        // Zuerst wird die "Lnge" der Grenze ermittelt.
        // Jede y-nderung dy der Grenze von x nach x+1 (und von x=511 nach x=0)
        // fhrt zu einem Punktabzug von dy/2048.
        // Negative Punktzahlen sind mglich!
        // Eine in dieser Hinsicht optimale Grenze ist einfach eine horizontale Linie.
        // Eine solche Grenze bringt genau 1 Punkt.

        int borderLength = 0;
        for(int x = 0; x < MAP_WIDTH - 1; ++x) borderLength += absInt(border[x] - border[x + 1]);
        borderLength += absInt(border[MAP_WIDTH - 1] - border[0]);
        const double borderScore = 1.0 - borderLength / 2048.0;

        // Nun wird die Qualitt der Aufteilung des Planeten bestimmt.
        // Diese besteht aus zwei Faktoren:
        //
        // 1.) Wasser(Wasserbewohner) / Wasser(Gesamt)
        //     = Anteil des nutzbaren Wassers am Gesamtvorkommen auf dem Planeten
        //       (Wasser, das den Landbewohnern zugeteilt wird, gilt als verschwendet)
        //
        // 2.) Land(Landbewohner) / Land(Gesamt)
        //     = Anteil des nutzbaren Lands am Gesamtvorkommen auf dem Planeten
        //       (Land, das den Wasserbewohnern zugeteilt wird, gilt als verschwendet)
        //
        // Diese Faktoren werden miteinander multipliziert
        // und ergeben die zweite Teilpunktzahl.
        // Das theoretische Maximum ist 1, nmlich wenn die eine Hlfte des Planeten
        // zu 100% aus Wasser besteht und die andere zu 100% aus Land.

        // bestimmen, wie viel Wasser/Land nrdlich und sdlich der Grenze vorkommen
        int waterNorth = 0;
        int waterSouth = 0;
        int landNorth = 0;
        int landSouth = 0;
        for(int x = 0; x < MAP_WIDTH; ++x)
        {
            // nrdliche Seite der Grenze
            int landNorthInColumn = 0;
            for(int y = 0; y < border[x]; ++y) landNorthInColumn += planetMap[x][y] ? 1 : 0;
            landNorth += landNorthInColumn;
            waterNorth += border[x] - landNorthInColumn;

            // sdliche Seite der Grenze
            int landSouthInColumn = 0;
            for(int y = border[x]; y < MAP_HEIGHT; ++y) landSouthInColumn += planetMap[x][y] ? 1 : 0;
            landSouth += landSouthInColumn;
            waterSouth += (MAP_HEIGHT - border[x]) - landSouthInColumn;
        }

        // Gesamtvorkommen berechnen
        const int waterTotal = waterNorth + waterSouth;
        const int landTotal = landNorth + landSouth;

        // Wir wissen noch nicht, welche Bewohner auf welcher Seite leben sollen.
        // Darum probieren wir beide Mglichkeiten und whlen dann die bessere.
        const double divisionScore1 = (waterNorth / static_cast<double>(waterTotal)) * (landSouth / static_cast<double>(landTotal));
        const double divisionScore2 = (waterSouth / static_cast<double>(waterTotal)) * (landNorth / static_cast<double>(landTotal));
        const double bestDivisionScore = std::max(divisionScore1, divisionScore2);

        // Die endgltige Punktzahl ergibt sich:
        // - zu 1/3 aus der Punktzahl fr die Lnge der Grenze
        // - zu 2/3 aus der Punktzahl fr die Qualitt der Aufteilung
        return (borderScore + 2.0 * bestDivisionScore) / 3.0;
    }

	/*
	 *	Mir ist bewusst, dass der nachfolgende Code nicht wirklich schn und bersichtlich
	 *	geschrieben worden ist, doch aufgrund von mangelnder Zeit habe ich es nicht besser
	 *	ausgearbeitet; aber immerhin, es funktioniert (mehr oder weniger^^).
	 */
	void dividePlanet(const PlanetMap& planetMap,
		Border& outBorder,
		const Clock& clock)
	{
		int count (0);
		int countMax (-9999);
		int countMin (9999);
		Border countMaxBorder;
		Border countMinBorder;
		int x, y;
		for (x=0; x<MAP_WIDTH; ++x)
		{
			// vorlufige Annahme: Sden = Wasser
			//
			// 1) Maximal und Minimalwerte berechnen
			for (y=MAP_HEIGHT-1; y>1; --y)
			{
				if (planetMap[x][y])    // Land
					--count;
				else                    // Wasser
					++count;
				if (count > countMax)
				{
					countMax = count;
					countMaxBorder[x] = y;
				}
				if (count < countMin)
				{
					countMin = count;
					countMinBorder[x] = y;
				}
			}

			// Aufrumarbeiten
			count = 0;
			countMax = -9999;
			countMin = 9999;
		}

		// 2) Grenze setzen
		if (computeScore(planetMap, countMinBorder) >= computeScore(planetMap, countMaxBorder))	// Min- oder Max-Werte besser? Wasser im Norden oder Sden?
			memcpy(outBorder, countMinBorder, sizeof(outBorder));								// --> Norden = Wasser, --> Min-Werte benutzen.
		else																					// --> Sden = Wasser, --> Max-Werte benutzen
			memcpy(outBorder, countMaxBorder, sizeof(outBorder));

		// 3) Interpolation
		Border tempBorder;
		double scoreOld = computeScore(planetMap, outBorder);

		while(clock() < 0.80f)
		{
			memcpy(tempBorder, outBorder, sizeof(tempBorder));

			for (int i = MAP_HEIGHT; i >= 1 && clock() < 0.80f; --i)
			{
				// 1. Punkt: (kein vorheriger Punkt)
				if (absInt(tempBorder[0] - tempBorder[1]) >= i )
				{
					tempBorder[0] = tempBorder[1];
				}
				// mittlere Punkte:
				for (int j = 1; j < MAP_WIDTH-1; ++j)
				{
					if (absInt(tempBorder[j] - tempBorder[j-1]) >= i ||		// Wenn der Abstand des aktuellen Punktes zum vorherigen bzw.
						absInt(tempBorder[j] - tempBorder[j+1]) >= i )		// nchsten zu gro ist, ...
					{
						int avg = (tempBorder[j-1] + tempBorder[j+1])/2;	// ... setze ihn in die Mitte beider Punkte.
						tempBorder[j] = avg;
					}
				}

				// letzter Punkt: (kein nachfolgender Punkt)
				if (absInt(tempBorder[MAP_WIDTH-1] - tempBorder[MAP_WIDTH-2]) >= i)
				{
					tempBorder[MAP_WIDTH-1] = tempBorder[MAP_WIDTH-2];
				}
			}

			double scoreNew = computeScore(planetMap, tempBorder);
			if (scoreNew > scoreOld)	// Besseres Ergebnis? Ja --> bernehmen.
			{
				memcpy(outBorder, tempBorder, sizeof(outBorder));
				scoreOld = scoreNew;
			}
			else						// Nein --> ~bestes Ergebnis bereits berechnet, Schleife abbrechen.
				break;
		}

		// 4) Spitzen begradigen --> krzere Grenze
		int maxDistance (2);
		memcpy(tempBorder, outBorder, sizeof(tempBorder));
		Border tempBorder2;
		while (clock() < 0.95f)
		{
			for (int x = 0; x < MAP_WIDTH && clock() < 0.95f; ++x)
			{
				int y = tempBorder[x];
				bool change = false;	// Erst bei mind. 1 Abweichung begradigen; sonst kein Effekt und nur Zeitverschwendung
				for(int i = x+1, j = 1; i < MAP_WIDTH && j <= maxDistance; ++i, ++j)
				{
					if (tempBorder[i] <= y+1 && tempBorder[i] >= y-1)	// Toleranz von +-1 Pixel
					{
						if (change)
						{
							memcpy(tempBorder2, tempBorder, sizeof(tempBorder2));
							for (int k = x; k <= i; ++k)
							{
								tempBorder2[k] = tempBorder2[x];
							}
							if (computeScore(planetMap, tempBorder2) > computeScore(planetMap, tempBorder))	// Nur wenn Verbesserung vorhanden, bernehmen
							{
								memcpy(tempBorder, tempBorder2, sizeof(tempBorder));
							}
							break;
						}
						else
							continue;
					}
					else
						change = true;
				}
			}

			// Wirklich besser?
			double scoreNew = computeScore(planetMap, tempBorder);
			if (scoreNew > scoreOld)	// Ja --> bernehmen
			{
				scoreOld = scoreNew;
				memcpy(outBorder, tempBorder, sizeof(outBorder));
			}

			if (maxDistance <= MAP_WIDTH/2)
				maxDistance *= 2;
			else
				break;

		}

		// 5) Alternative?
		for (int i = 1; i < MAP_HEIGHT && clock() < 0.99f; ++i)		// Ist eine gerade Linie vielleicht besser als der obere Algorithmus? berprfen.
		{															// (--> Teil 1 des oberen Algorithmus verbesserungswrdig)
			for (int j = 0; j < MAP_WIDTH; ++j)
			{
				tempBorder[j] = i;
			}
			double scoreNew = computeScore(planetMap, tempBorder);
			if (scoreNew > scoreOld)	// Gerade Linie scheint besser zu sein..., bernehmen
			{
				scoreOld = scoreNew;
				memcpy(outBorder, tempBorder, sizeof(outBorder));
			}
		}

	}

}	// namespace Aremion
