namespace Helmut
{
//Der Algo ist eigentlich recht einfach. Wir starten mit der Grenze auf dem quator
//und setzen dann die Grenze probeweise von x1 bis x2 auf einen neuen Wert y.
//Wir berechnen alle Kombinationen, also (MAP_WIDTH^2 * MAP_HEIGHT) neue Grenzen.
//Von all diesen Grenzen fischen wir die heraus, die die beste Wertung liefert und
//wiederholen mit dieser das ganze Spiel, bis keine nderung mehr eine bessere Wertung
//liefert.
//
//Der Trick ist nun, das ganze in einer Sekunde zu schaffen. Zum einen erstelle ich
//einige Lookuptabellen, mit denen man den neuen Score einer Grenze mit einigen
//wenigen Additionen komplett ohne Schleife ausrechnen kann (egal wie weit x1 und
//x2 auseinander liegen).
//Zum anderen werden nicht in jedem Durchgang alle mglichen neuen Grenzen
//ausgerechnet, sondern zB nur alle x1 Werte, die durch 10 teilbar sind. Vom besten
//x1 Wert werden dann alle umliegenden Werte berechnet.
//
//Auf meinem Computer braucht der Algo etwa 2 Sekunden pro Karte, um eine "perfekte"
//Grenze zu liefern. Also eine Grenze, bei der es garantiert kein Wertetripel (x1,x2,y)
//gibt, mit dem man die Grenze verbessern knnte. Aber nach einer Sekunde liefert der
//Algo auch schon sehr gute Ergebnisse. In den 100 Testkarten liegt der Algo nach einer
//Sekunde im Durchschnitt nur ~0,000003 Punkte pro Karte hinter dem "perfekten"
//Punktestand zurck. Mit je einer Sekunde Rechenzeit pro Karte erzielt der Algo auf
//meinem Rechner 76.4855 Punkte fr die 100 Testkarten.
//Es kann natrlich vorkommen, dass die Grenze trotzdem noch mit zwei oder mehr
//solchen Tripeln verbessert werden kann. In Planet000 zB gert der Algo in ein lokales
//Maximum. Dafr hab ich den Algo etwas erweitert, sodass im Bereich x1 bis x2 nochmal
//die halbe Grenze verschoben wird, wenn die totale Grenzlnge dadurch nicht beeinflusst
//wird, um die Flchenausnutzung noch weiter zu optimieren.

void dividePlanet(const PlanetMap& planetMap,
                  Border& border,
                  const Clock& clock)
{
//#define WINDOW_OUTPUT
//Zeigt die Zwischenschritte in einem Fenster. Rendern ist sehr langsam,
//also besser die Uhr verlangsamen, weil der Algo sonst zu frh abbricht.
//Dafr zB in der Klasse Clock:
//"conv(1.0 / CLOCKS_PER_SEC)" in "conv(.001 / CLOCKS_PER_SEC)" ndern

#ifdef WINDOW_OUTPUT
	static HWND hwnd = 0;
	int xFix = -1, yFix = -1;
	if(hwnd == 0)
	{
		WNDCLASSW wndclass;
		wndclass.style         = CS_HREDRAW | CS_VREDRAW;
		wndclass.lpfnWndProc   = DefWindowProcW;
		wndclass.cbClsExtra    = 0;
		wndclass.cbWndExtra    = 0;
		wndclass.hInstance     = GetModuleHandle(0);
		wndclass.hIcon         = LoadIcon(0, IDI_APPLICATION);
		wndclass.hCursor       = LoadCursor(0, IDC_ARROW);
		wndclass.hbrBackground = (HBRUSH)GetStockObject(NULL_BRUSH);
		wndclass.lpszMenuName  = 0;
		wndclass.lpszClassName = L"Planetenteilung";

		RegisterClassW(&wndclass);

		RECT rc;
		SetRect(&rc, 0, 0, MAP_WIDTH, MAP_HEIGHT);

		AdjustWindowRectEx(&rc, WS_OVERLAPPEDWINDOW, false, 0);

		hwnd = CreateWindowW(L"Planetenteilung", L"",
			WS_OVERLAPPEDWINDOW | WS_VISIBLE,
			400, 200,
			rc.right-rc.left, rc.bottom-rc.top,
			0, 0, GetModuleHandle(0), 0);
	}
#endif

//	Wir brauchen erstmal rund 3 MB Speicher. Leider ist der Stack dafr
//	normalerweise zu klein und global machen ist verboten... Also
//	vom Heap holen und 2D Arrays mit Makros hinwurschteln...

/*	int waterUsedLookup[MAP_WIDTH][MAP_HEIGHT];
	int landUsedLookup[MAP_WIDTH][MAP_HEIGHT];

	int waterUsedChangeLookup[MAP_WIDTH][MAP_HEIGHT];
	int landUsedChangeLookup[MAP_WIDTH][MAP_HEIGHT];

	int waterUsedChangeLookupSum[MAP_WIDTH+1][MAP_HEIGHT];
	int landUsedChangeLookupSum[MAP_WIDTH+1][MAP_HEIGHT];

	int borderLookupSum[MAP_WIDTH+1];*/

	int* waterUsedLookup_ = new int[MAP_WIDTH*MAP_HEIGHT + //allokieren wir's in einem Rutsch
									MAP_WIDTH*MAP_HEIGHT +
									MAP_WIDTH*MAP_HEIGHT +
									MAP_WIDTH*MAP_HEIGHT +
									(MAP_WIDTH+1)*MAP_HEIGHT +
									(MAP_WIDTH+1)*MAP_HEIGHT +
									MAP_WIDTH+1];
	int* landUsedLookup_ = waterUsedLookup_ + MAP_WIDTH*MAP_HEIGHT;

	int* waterUsedChangeLookup_ = landUsedLookup_ + MAP_WIDTH*MAP_HEIGHT;
	int* landUsedChangeLookup_ = waterUsedChangeLookup_ + MAP_WIDTH*MAP_HEIGHT;

	int* waterUsedChangeLookupSum_ = landUsedChangeLookup_ + MAP_WIDTH*MAP_HEIGHT;
	int* landUsedChangeLookupSum_ = waterUsedChangeLookupSum_ + (MAP_WIDTH+1)*MAP_HEIGHT;

	int* borderLookupSum = landUsedChangeLookupSum_ + (MAP_WIDTH+1)*MAP_HEIGHT;

#define waterUsedLookup(x,y) waterUsedLookup_[(x)*MAP_HEIGHT+(y)]
#define landUsedLookup(x,y) landUsedLookup_[(x)*MAP_HEIGHT+(y)]
#define waterUsedChangeLookup(x,y) waterUsedChangeLookup_[(x)*MAP_HEIGHT+(y)]
#define landUsedChangeLookup(x,y) landUsedChangeLookup_[(x)*MAP_HEIGHT+(y)]
#define waterUsedChangeLookupSum(x,y) waterUsedChangeLookupSum_[(x)*MAP_HEIGHT+(y)]
#define landUsedChangeLookupSum(x,y) landUsedChangeLookupSum_[(x)*MAP_HEIGHT+(y)]


	//Das Datenformat von planetMap ist etwas unhandlich.
	//Wir konvertieren es in zwei 2D Arrays, die beschreiben,
	//wie viel Flche genutzt wird, wenn die Grenze auf eine
	//bestimmte Stelle gesetzt wird. (planetMap wird danach
	//nicht mehr benutzt)

	for(int x = 0; x < MAP_WIDTH; ++x)
	{
		waterUsedLookup(x,0) = 0;
		for(int y = 1; y < MAP_HEIGHT; ++y)
			waterUsedLookup(x,y) = waterUsedLookup(x,y-1) + (planetMap[x][y-1] ? 0 : 1);
	}
	for(int x = 0; x < MAP_WIDTH; ++x)
	{
		landUsedLookup(x,MAP_HEIGHT-1) = planetMap[x][MAP_HEIGHT-1] ? 1 : 0;
		for(int y = MAP_HEIGHT-2; y >= 0; --y)
			landUsedLookup(x,y) = landUsedLookup(x,y+1) + (planetMap[x][y] ? 1 : 0);
	}

	int waterTotal = 0;
	for(int x = 0; x < MAP_WIDTH; ++x)
		waterTotal += MAP_HEIGHT - landUsedLookup(x,0);//das geht wirklich;)

	//Jetzt legen wir durch ne Schtzung fest, welche Aliens im Norden/Sden leben sollen.
	//Spter ist kein Wechsel mehr mglich, deshalb ist die Heuristik ziemlich aufwendig.
	//In den 100 Testkarten liegt sie nur etwa 3 Mal falsch, und auch nur bei Karten wo es
	//keinen groen Punktunterschied ausmacht.
	double waterFactor = 1.0/double(waterTotal);
	double landFactor = 1.0/double(MAP_WIDTH*MAP_HEIGHT-waterTotal);

	double cUsedNotSwap = 0.0, cUsedSwap = 0.0;

	for(int x = 0; x < MAP_WIDTH; ++x)
	{
		double bestDivisionNotSwap = -1000;
		double bestDivisionSwap = -1000;
		for(int y = 1; y < MAP_HEIGHT; ++y)
		{
			double divisionNotSwap = waterUsedLookup(x,y)*waterFactor + landUsedLookup(x,y)*landFactor;
			if(divisionNotSwap > bestDivisionNotSwap)
			{
				bestDivisionNotSwap = divisionNotSwap;
			}
			double divisionSwap = (y - waterUsedLookup(x,y))*landFactor + (MAP_HEIGHT-y-landUsedLookup(x,y))*waterFactor;
			if(divisionSwap > bestDivisionSwap)
			{
				bestDivisionSwap = divisionSwap;
			}
		}
		cUsedNotSwap += bestDivisionNotSwap;
		cUsedSwap += bestDivisionSwap;
	}
	if(cUsedSwap > cUsedNotSwap)
	{
		//Wasseraliens lieber unten ansiedeln.
		//Wir tauschen dafr einfach Land und Wasserflchen in
		//waterUsedLookup und landUsedLookup aus
		for(int x = 0; x < MAP_WIDTH; ++x)
		for(int y = 0; y < MAP_HEIGHT; ++y)
		{
			waterUsedLookup(x,y) = y - waterUsedLookup(x,y);
			landUsedLookup(x,y) = MAP_HEIGHT - y - landUsedLookup(x,y);
		}
		waterTotal = MAP_WIDTH * MAP_HEIGHT - waterTotal;
	}
	//Jetzt knnen wir davon ausgehen, dass die Wasseraliens
	//im Norden leben

	const int landTotal = MAP_WIDTH * MAP_HEIGHT - waterTotal;

	//Mit Mittellinie anfangen
	for(int x = 0; x < MAP_WIDTH; ++x)
		border[x] = MAP_HEIGHT/2;

	const int maxDXPerPass[] = {MAP_WIDTH, 439, 239, 172, 155, 130, 126, 113, 303, 117, 83, 34};

	int step = 26;
	for(int pass = 0;; pass++)
	{
#ifdef WINDOW_OUTPUT
		if(GetAsyncKeyState('P') >= 0)
		{
			wchar_t Buffer[512];
			_snwprintf(Buffer, 512, L"Planetenteilung - Helmut (Durchgang: %02i  Score: %f)", pass, computeScore(planetMap, border));
			Buffer[512-1] = 0;
			SetWindowTextW(hwnd, Buffer);

			POINT pt;
			GetCursorPos(&pt);
			ScreenToClient(hwnd, &pt);
			if(GetForegroundWindow() == hwnd &&
				GetAsyncKeyState(VK_LBUTTON) < 0 &&
				pt.x >= 0 && pt.x < MAP_WIDTH &&
				pt.y >= 1 && pt.y < MAP_HEIGHT)
			{
				xFix = pt.x;
				yFix = pt.y;
				border[xFix] = yFix;
			}
			else
				xFix = yFix = -1;

			HDC hdc = GetDC(hwnd);

			for(int x = 0; x < MAP_WIDTH; ++x)
			{
				for(int y = 0; y < MAP_HEIGHT; ++y)
				{
					if(y >= border[x])
						SetPixel(hdc, x, y, planetMap[x][y] ? RGB(255,100,0) : RGB(100,0,255));
					else
						SetPixel(hdc, x, y, planetMap[x][y] ? RGB(255,255,0) : RGB(0,0,255));
				}
			}
			ReleaseDC(hwnd, hdc);
		}
#endif

		if(step > 1)//Mit der Zeit genauer werden
			step--;

		//Ein paar Lookup-Tables berechnen
		//Diese hngen alle von border ab und mssen deshalb immer wieder neu berechnet werden
		for(int x = 0; x < MAP_WIDTH; ++x)
		{
			int prevWaterUsed = waterUsedLookup(x,border[x]);
			int prevLandUsed = landUsedLookup(x,border[x]);
			for(int y = 0; y < MAP_HEIGHT; ++y)
			{
				waterUsedChangeLookup(x,y) = waterUsedLookup(x,y) - prevWaterUsed;
				landUsedChangeLookup(x,y) =  landUsedLookup(x,y) - prevLandUsed;
			}
		}

		//3 Lookuptables hintereinandergeschaltet. Das tat echt weh sich sowas auszudenken...
		for(int y = 0; y < MAP_HEIGHT; ++y)
		{
			waterUsedChangeLookupSum(0,y) = 0;
			landUsedChangeLookupSum(0,y) = 0;
			for(int x = 1; x < MAP_WIDTH+1; ++x)
			{
				waterUsedChangeLookupSum(x,y) = waterUsedChangeLookupSum(x-1,y) + waterUsedChangeLookup(x-1,y);
				landUsedChangeLookupSum(x,y) = landUsedChangeLookupSum(x-1,y) + landUsedChangeLookup(x-1,y);
			}
		}

		borderLookupSum[0] = 0;
		for(int x = 1; x < MAP_WIDTH+1; ++x)
			borderLookupSum[x] = borderLookupSum[x-1] + absInt(border[x-1]-border[x%MAP_WIDTH]);

		//Aktuellen Score berechnen
		int waterNorth = 0;
		int landSouth = 0;
		for(int x = 0; x < MAP_WIDTH; ++x)
		{
			waterNorth += waterUsedLookup(x,border[x]);
			landSouth += landUsedLookup(x,border[x]);
		}

        //Blder GCC optimiert double Rechnungen, sodass sich das Ergebnis ndert, wenn
        //man diese Zwischenschritte nicht volatile macht.
		volatile double divisionScore = (waterNorth / static_cast<double>(waterTotal)) * (landSouth / static_cast<double>(landTotal));
        volatile double borderScore = (1.0 - borderLookupSum[MAP_WIDTH] / 2048.0);
        volatile double score = (borderScore + 2.0 * divisionScore) / 3.0;
        //assert(score == computeScore(planetMap, border));

		//Neue (bessere) Grenze berechnen
		//Da die Lookuptables mit jeder Grenzennderung ungltig werden drften wir
		//eigentlich nur eine nderung pro Durchlauf durchfhren.
		//Wir speichern aber trotzdem einfach so viele wie mglich, das geht halt schneller.
		//Allerdings sorgen wir dafr, dass bessere Grenzennderungen sich immer durchsetzen
		//und jede nderung entweder ganz oder garnicht durchgefhrt wird.
		//Sonst wrden sich viele nderungen gegenseitig ausheben und den Score so
		//u.U. sogar verringern.
		struct BorderItem
		{
			double score;
			int y;
			int xStart, xEnd;
		};
		BorderItem newBorder[MAP_WIDTH];
		for(int x = 0; x < MAP_WIDTH; ++x)
		{
			newBorder[x].score = 0.0;
			newBorder[x].y = border[x];
			newBorder[x].xStart = newBorder[x].xEnd = -1;
		}

        //bestScore gibt an, auf wie viel wir den aktuellen Score verbessern knnen (hiernach maximieren wir)
		double bestScore = score;
//		int bestX, bestDX, bestY;

		for(int x = 0; x < MAP_WIDTH; x++)
		{
			if(clock() >= 0.995)
				break;

#ifdef WINDOW_OUTPUT
			MSG msg;
			while(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
			{
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
#endif


			int dxStep = std::max(step, 1);//Wir checken erst grob, beim besten Treffer schauen wir dann genau
			int dxEnd = pass >= int(sizeof(maxDXPerPass)/sizeof(maxDXPerPass[0])) ? 5 : maxDXPerPass[pass]+1;
			double dxBestScore = -1000.0;
			int dxBestStart = 1;
			for(int dx = dxStep; dx < dxEnd; dx += dxStep)
			{
				//assert(dx >= 1 && dx <= MAP_WIDTH);
				int x2 = (x+dx)%MAP_WIDTH;
				int xLast = (x+dx-1)%MAP_WIDTH;

				int middleBorderDelta;
				if(dx == MAP_WIDTH)
				{
					//verschieben wir die komplette Grenze?
					middleBorderDelta = -borderLookupSum[MAP_WIDTH];
				}
				else if(xLast >= x)
					middleBorderDelta = -(borderLookupSum[xLast] - borderLookupSum[x]);
				else
					middleBorderDelta = -(borderLookupSum[xLast] + borderLookupSum[MAP_WIDTH] - borderLookupSum[x]);

				int yStep = step == 0 || pass > 13 ? 1 : 7;//ist optimierter Code nicht wunderbar?
				int yEnd = MAP_HEIGHT;
				double yBestScore = -1000.0;
				int yBestStart = 1;
				for(int y = yStep; y < yEnd; y += yStep)
				{
					//Neuen Score fr die Border berechnen, wenn man
					//die Grenze von x bis x2 auf y setzt
					//assert(y >= 1 && y < MAP_HEIGHT);

#ifdef WINDOW_OUTPUT
					if(xFix != -1)
					{
						if(xLast >= x)
						{
							if(x <= xFix && xFix <= xLast && yFix != y) continue;
						}
						else
						{
							if((x <= xFix || xFix <= xLast) && yFix != y) continue;
						}
						if(dxStep != 1) dxEnd = MAP_WIDTH;
					}
#endif

					//Wie ndert sich die Grenzlnge?
					int leftBorder = border[(x-1+MAP_WIDTH)%MAP_WIDTH];
					int leftBorderDelta = absInt(y-leftBorder) - absInt(border[x]-leftBorder);

					int rightBorderDelta = absInt(y-border[x2]) - absInt(border[x2]-border[xLast]);
					if(dx == MAP_WIDTH)
					{
						//verschieben wir die komplette Grenze?
						leftBorderDelta = rightBorderDelta = 0;
					}

					int newBorderLength = borderLookupSum[MAP_WIDTH] + leftBorderDelta + rightBorderDelta + middleBorderDelta;

					//Wie ndert sich die genutzte Flche?
					int waterDelta, landDelta;
					if(x2 > x)
					{
						waterDelta = waterUsedChangeLookupSum(x2,y) - waterUsedChangeLookupSum(x,y);
						landDelta =  landUsedChangeLookupSum(x2,y) -  landUsedChangeLookupSum(x,y);
					}
					else
					{
						waterDelta = waterUsedChangeLookupSum(x2,y) + waterUsedChangeLookupSum(MAP_WIDTH,y) - waterUsedChangeLookupSum(x,y);
						landDelta =  landUsedChangeLookupSum(x2,y) +  landUsedChangeLookupSum(MAP_WIDTH,y) -  landUsedChangeLookupSum(x,y);
					}

					//In die Mitte der neuen Grenze die Hhe ndern, falls es die
					//Grenzlnge nicht verndert und die Flchenausnutzung dadurch
					//verbessert wird
					int y2 = -1;
					int xMiddle = (x+dx/2)%MAP_WIDTH;
					if(pass >= 5 && dx >= 2 && std::min(border[x2], leftBorder) <= y && y <= std::max(border[x2], leftBorder))
					{
						int y2Min = std::min(border[x2]+1, y+1);
						int y2Max = std::max(border[x2]-1, y-1);
						if(y2Max-y2Min < 20)
						{
							int waterDeltaBase, landDeltaBase;
							if(xMiddle > x)
							{
								waterDeltaBase = waterUsedChangeLookupSum(xMiddle,y) - waterUsedChangeLookupSum(x,y);
								landDeltaBase =  landUsedChangeLookupSum(xMiddle,y) -  landUsedChangeLookupSum(x,y);
							}
							else
							{
								waterDeltaBase = waterUsedChangeLookupSum(xMiddle,y) + waterUsedChangeLookupSum(MAP_WIDTH,y) - waterUsedChangeLookupSum(x,y);
								landDeltaBase =  landUsedChangeLookupSum(xMiddle,y) +  landUsedChangeLookupSum(MAP_WIDTH,y) -  landUsedChangeLookupSum(x,y);
							}
							for(int yy = y2Min; yy <= y2Max; yy++)
							{
								int waterDelta2, landDelta2;
								if(x2 > xMiddle)
								{
									waterDelta2 = waterUsedChangeLookupSum(x2,yy) - waterUsedChangeLookupSum(xMiddle,yy);
									landDelta2 =  landUsedChangeLookupSum(x2,yy) -  landUsedChangeLookupSum(xMiddle,yy);
								}
								else
								{
									waterDelta2 = waterUsedChangeLookupSum(x2,yy) + waterUsedChangeLookupSum(MAP_WIDTH,yy) - waterUsedChangeLookupSum(xMiddle,yy);
									landDelta2 =  landUsedChangeLookupSum(x2,yy) +  landUsedChangeLookupSum(MAP_WIDTH,yy) -  landUsedChangeLookupSum(xMiddle,yy);
								}
								if(//waterDeltaBase+waterDelta2 > waterDelta && landDeltaBase+landDelta2 > landDelta)
									((waterNorth+waterDeltaBase+waterDelta2) / static_cast<double>(waterTotal)) * ((landSouth+landDeltaBase+landDelta2) / static_cast<double>(landTotal)) >
									((waterNorth+waterDelta                ) / static_cast<double>(waterTotal)) * ((landSouth+landDelta               ) / static_cast<double>(landTotal)))
								{
									waterDelta = waterDeltaBase+waterDelta2;
									landDelta = landDeltaBase+landDelta2;
									y2 = yy;
								}
							}
						}
					}
					if(y2 == -1)
						xMiddle = -1;

					//Score der neuen Grenze berechnen
                    volatile double newBorderScore = (1.0 - newBorderLength / 2048.0);
					volatile double newDivisionScore = ((waterNorth+waterDelta) / static_cast<double>(waterTotal)) * ((landSouth+landDelta) / static_cast<double>(landTotal));
                    volatile double newScore = (newBorderScore + 2.0 * newDivisionScore) / 3.0;
					if(newScore > bestScore)
					{
						bestScore = newScore;
//						bestX = x;
//						bestX2 = x2;
//						bestDX = dx;
//						bestY = y;
//						bestBorderDelta = leftBorderDelta + rightBorderDelta + middleBorderDelta;
//						bestWaterDelta = waterDelta;
					}
					if(newScore > yBestScore)
					{
						yBestScore = newScore;
						yBestStart = y;
					}
					if(newScore > dxBestScore)
					{
						dxBestScore = newScore;
						dxBestStart = dx;
					}
					if(newScore >= score)
					{
						//Ist unsere neue Border auch an allen Stellen
						//besser als die jeweils vorherige?
						bool insert = true;
						for(int xx = x;;)
						{
							if(newScore < newBorder[xx].score)
							{
								insert = false;
								break;
							}
							xx = (xx+1)%MAP_WIDTH;
							if(xx == x2) break;//nicht oben in Schleife, damit im ersten Durchlauf nicht schon abgebrochen wird wenn x==x2 ist
						}
						if(insert)
						{
//							if(dx > MAX_DX)
//								MAX_DX = dx;
							int newY = y;
							for(int xx = x;;)
							{
								//bisherige Border lschen
								int xStart = newBorder[xx].xStart;
								int xEnd = newBorder[xx].xEnd;
								if(xStart != -1)
									for(int xxx = xStart;;)
									{
										newBorder[xxx].score = 0.0;
										newBorder[xxx].y = border[xxx];
										newBorder[xxx].xStart = newBorder[xxx].xEnd = -1;
										xxx = (xxx+1)%MAP_WIDTH;
										if(xxx == xEnd) break;
									}
								//assert(newBorder[xx].xStart == -1);
								//assert(newBorder[xx].score == 0.0);
								newBorder[xx].score = newScore;
								if(xx == xMiddle)
									newY = y2;
								newBorder[xx].y = newY;
								newBorder[xx].xStart = x;
								newBorder[xx].xEnd = x2;

								xx = (xx+1)%MAP_WIDTH;
								if(xx == x2) break;
							}
						}
					}

					if(yStep != 1 && y+yStep >= yEnd)
					{
						y = std::max(yBestStart-yStep, 0);
						yEnd = std::min(yBestStart+yStep, MAP_HEIGHT);
						yStep = 1;
					}
				}
				if(dxStep != 1 && dx+dxStep >= dxEnd)
				{
					if(x == 0 && dx != MAP_WIDTH && pass == 0)
						dx = MAP_WIDTH-dxStep;
					else
					{
						dx = std::max(dxBestStart-dxStep, 0);
						dxEnd = std::min(dxBestStart+dxStep, MAP_WIDTH);
						dxStep = 1;
					}
				}
			}
		}

		//std::cout << "bestScore: " << bestScore << " " << bestX << " "  << bestDX << " "  << bestY << " " << std::endl;

		if(bestScore == score)
		{
			//Keine Verbesserung gefunden :(
			if(step != 0)
			{
				//Nochmal ganz genau suchen, solange wir Zeit haben
				step = 0;
			}
			else
			{
				//Nichts mehr zu machen, das Ergebnis drfte ziemlich perfekt sein
//				std::cout << "Passes: " << passes << " ";
				break;
			}
		}
		else
		{
			for(int x = 0; x < MAP_WIDTH; ++x)
				border[x] = newBorder[x].y;
		}
		if(clock() >= 0.995)
		{
//			std::cout << "Passes: " << passes << " ";
			break;
		}
	}

	delete[] waterUsedLookup_;

#undef waterUsedLookup
#undef landUsedLookup
#undef waterUsedChangeLookup
#undef landUsedChangeLookup
#undef waterUsedChangeLookupSum
#undef landUsedChangeLookupSum
}

/*
//Ergebnis auf meinem Rechner:

Planet         Zeit (ms)  Score
----------------------------------
Planet000.png  990        0.7834221335
Planet001.png  865        0.7961629983
Planet002.png  990        0.7726642929
Planet003.png  990        0.7274000954
Planet004.png  990        0.7637908064
Planet005.png  990        0.6675136974
Planet006.png  995        0.8596858993
Planet007.png  992        0.7030496437
Planet008.png  993        0.7738361927
Planet009.png  990        0.8386284744
Planet010.png  993        0.7309861404
Planet011.png  990        0.6624861413
Planet012.png  990        0.7844969985
Planet013.png  808        0.6427255215
Planet014.png  990        0.8376783229
Planet015.png  990        0.8110677211
Planet016.png  990        0.6172354704
Planet017.png  995        0.6823663013
Planet018.png  990        0.7352391638
Planet019.png  990        0.6806118778
Planet020.png  995        0.8665153187
Planet021.png  990        0.7662221268
Planet022.png  990        0.6441223564
Planet023.png  990        0.8926346949
Planet024.png  990        0.8816112167
Planet025.png  990        0.9245866283
Planet026.png  990        0.7678285068
Planet027.png  990        0.7496575664
Planet028.png  990        0.5675424351
Planet029.png  990        0.7793545127
Planet030.png  990        0.6714112072
Planet031.png  992        0.9426993357
Planet032.png  996        0.7468501219
Planet033.png  990        0.9200866899
Planet034.png  990        0.6798032863
Planet035.png  990        0.8039539236
Planet036.png  990        0.9044127986
Planet037.png  990        0.7563892764
Planet038.png  993        0.6529661851
Planet039.png  990        0.8870204075
Planet040.png  990        0.6996330964
Planet041.png  990        0.8835569627
Planet042.png  990        0.6812422949
Planet043.png  990        0.8017103111
Planet044.png  990        0.6629237487
Planet045.png  990        0.6738006197
Planet046.png  990        0.7073334419
Planet047.png  990        0.5732543645
Planet048.png  990        0.8529707863
Planet049.png  990        0.6401970684
Planet050.png  990        0.7143449956
Planet051.png  995        0.9369923343
Planet052.png  990        0.8570513264
Planet053.png  990        0.6468187281
Planet054.png  990        0.6787310482
Planet055.png  990        0.8315303318
Planet056.png  990        0.7104426595
Planet057.png  990        0.7592426646
Planet058.png  990        0.7970944522
Planet059.png  990        0.7589535234
Planet060.png  990        0.7739049225
Planet061.png  990        0.7205446054
Planet062.png  990        0.833043625
Planet063.png  990        0.8064390907
Planet064.png  990        0.8434206005
Planet065.png  990        0.6437249316
Planet066.png  955        0.6843563349
Planet067.png  990        0.6278633425
Planet068.png  990        0.871362638
Planet069.png  990        0.7757320264
Planet070.png  813        0.7556650118
Planet071.png  995        0.8085052767
Planet072.png  992        0.8257193568
Planet073.png  990        0.8487640906
Planet074.png  990        0.6901569688
Planet075.png  992        0.8061257005
Planet076.png  990        0.8757291554
Planet077.png  992        0.7771279339
Planet078.png  990        0.819245021
Planet079.png  990        0.8113572125
Planet080.png  995        0.6901238528
Planet081.png  990        0.7117609283
Planet082.png  990        0.8342389841
Planet083.png  990        0.7997136723
Planet084.png  990        0.8100583247
Planet085.png  990        0.7285309302
Planet086.png  993        0.6001526157
Planet087.png  993        0.805638609
Planet088.png  990        0.8382629739
Planet089.png  990        0.7605559686
Planet090.png  990        0.7884534126
Planet091.png  992        0.6951409186
Planet092.png  990        0.7455521491
Planet093.png  990        0.7023057422
Planet094.png  990        0.8048674655
Planet095.png  990        0.8232596363
Planet096.png  990        0.6825947006
Planet097.png  992        0.8422038894
Planet098.png  991        0.836222226
Planet099.png  993        0.8425485319

Gesamtpunktzahl: 76.48551462 Zeit: 98.55 s


Wenn man gengend Zeit lsst:

Planet         Zeit (ms)/100  Score
----------------------------------
Planet000.png  17         0.7834221335
Planet001.png  8          0.7961629983
Planet002.png  16         0.7726642929
Planet003.png  14         0.7274000954
Planet004.png  13         0.7637908064
Planet005.png  15         0.6675136974
Planet006.png  32         0.8596858993
Planet007.png  35         0.7030496437
Planet008.png  46         0.7738361927
Planet009.png  14         0.8386284744
Planet010.png  43         0.7309869561
Planet011.png  13         0.6624861413
Planet012.png  10         0.7844969985
Planet013.png  9          0.6427255215
Planet014.png  14         0.8376783229
Planet015.png  16         0.8110677211
Planet016.png  15         0.6172354704
Planet017.png  37         0.6823663013
Planet018.png  15         0.7352391638
Planet019.png  24         0.6806175179
Planet020.png  31         0.8665153187
Planet021.png  12         0.7662221268
Planet022.png  14         0.6441223564
Planet023.png  32         0.8926346949
Planet024.png  9          0.8816112167
Planet025.png  10         0.9245866283
Planet026.png  13         0.7678285068
Planet027.png  12         0.7496575664
Planet028.png  11         0.5675424351
Planet029.png  25         0.7793691202
Planet030.png  15         0.6714112072
Planet031.png  26         0.9426993357
Planet032.png  32         0.7468501219
Planet033.png  12         0.9200866899
Planet034.png  21         0.679814171
Planet035.png  13         0.8039539236
Planet036.png  39         0.9044127986
Planet037.png  21         0.7563892764
Planet038.png  33         0.6529661851
Planet039.png  13         0.8870204075
Planet040.png  17         0.6996446512
Planet041.png  17         0.8835569627
Planet042.png  15         0.6812425141
Planet043.png  15         0.801715615
Planet044.png  16         0.6629237487
Planet045.png  19         0.6738043714
Planet046.png  52         0.7074415493
Planet047.png  32         0.5732543645
Planet048.png  16         0.8529757137
Planet049.png  12         0.6401970684
Planet050.png  16         0.7143449956
Planet051.png  31         0.9369923343
Planet052.png  40         0.8570513264
Planet053.png  15         0.6468187281
Planet054.png  14         0.6787310482
Planet055.png  32         0.8315303318
Planet056.png  15         0.7104426595
Planet057.png  14         0.7592426646
Planet058.png  24         0.7970944532
Planet059.png  20         0.7589535234
Planet060.png  17         0.7739049225
Planet061.png  14         0.7205446054
Planet062.png  50         0.8330842
Planet063.png  16         0.8064390907
Planet064.png  13         0.8434206005
Planet065.png  15         0.6437324571
Planet066.png  10         0.6843563349
Planet067.png  15         0.6278987599
Planet068.png  13         0.871362638
Planet069.png  15         0.7757320264
Planet070.png  8          0.7556650118
Planet071.png  39         0.8085121601
Planet072.png  44         0.8257193568
Planet073.png  12         0.8487640906
Planet074.png  11         0.6901623112
Planet075.png  43         0.8061425642
Planet076.png  12         0.8757291554
Planet077.png  33         0.7771279339
Planet078.png  13         0.819245021
Planet079.png  12         0.8113572125
Planet080.png  30         0.6901238528
Planet081.png  9          0.7117609283
Planet082.png  15         0.8342389841
Planet083.png  15         0.7997154453
Planet084.png  10         0.8100583247
Planet085.png  8          0.7285309302
Planet086.png  32         0.6001526157
Planet087.png  48         0.805638609
Planet088.png  15         0.8382629739
Planet089.png  23         0.7605589701
Planet090.png  13         0.7884534126
Planet091.png  43         0.6951653442
Planet092.png  15         0.745562492
Planet093.png  9          0.7023057422
Planet094.png  15         0.8048674655
Planet095.png  15         0.8232596363
Planet096.png  36         0.6825947006
Planet097.png  32         0.8422038894
Planet098.png  12         0.836222226
Planet099.png  42         0.8425531683

Gesamtpunktzahl: 76.48583722 Zeit: 212.515 s

*/

// --------------------------------------------------------------------------------
}
