#include <stdio.h>
#include <string.h>
#include <stdlib.h>

namespace Wuempftlbruempftl
{

// needs access in struct method
int landTotal;
int waterTotal;
int landPerX[MAP_WIDTH];
int waterPerX[MAP_WIDTH];
void dividePlanet(const PlanetMap& planetMap, Border& outBorder, const Clock& clock)
{
	/////////////////////////////////////////////////////////////////////
	// FAST TRUE SCORE CALCULATOR
	/////////////////////////////////////////////////////////////////////
	struct FastBorder
	{
		FastBorder(const PlanetMap& _planetMap) : planetMap(_planetMap) {}

		// only a specific area changed:
		double computeScore(int dirtyStart, int dirtyEnd)
		{
		/*	// validate
			for(; x < MAP_WIDTH; ++x)
				if(border[x] <= 0 || border[x] >= MAP_HEIGHT)
				   throw std::runtime_error("Ungltiger Grenzverlauf!"); */

			int x = 0;
			unsigned int borderLength = 0;
			for(x = 0; x < MAP_WIDTH - 1; ++x)
				borderLength += absInt(border[x] - border[x + 1]);		// bit operation remove sign?
			borderLength += absInt(border[MAP_WIDTH - 1] - border[0]);
			const double borderScore = 1.0f - borderLength / 2048.0f;

			unsigned int waterNorth = 0;
			unsigned int waterSouth = 0;
			unsigned int landNorth = 0;
			unsigned int landSouth = 0;
			x = 0;
			for(; x < dirtyStart; ++x)
			{
				// north
				landNorth += landNorthColumn[x];
				unsigned int waterNorthColumn = border[x] - landNorthColumn[x];
				waterNorth += waterNorthColumn;
				// south
				waterSouth += waterPerX[x] - waterNorthColumn;
				landSouth  += landPerX[x] - landNorthColumn[x];
			}

			for(; x < dirtyEnd; ++x)
			{
				// north
				landNorthColumn[x] = 0;
				for(int y = 0; y < border[x]; ++y)
					landNorthColumn[x] += planetMap[x][y] ? 1 : 0;
				landNorth += landNorthColumn[x];
				unsigned int waterNorthColumn = border[x] - landNorthColumn[x];
				waterNorth += waterNorthColumn;
				// south
				waterSouth += waterPerX[x] - waterNorthColumn;
				landSouth  += landPerX[x] - landNorthColumn[x];
			}
			for(; x < MAP_WIDTH; ++x)
			{
				// north
				landNorth += landNorthColumn[x];
				unsigned int waterNorthColumn = border[x] - landNorthColumn[x];
				waterNorth += waterNorthColumn;
				// south
				waterSouth += waterPerX[x] - waterNorthColumn;
				landSouth  += landPerX[x] - landNorthColumn[x];
			}
			const double divisionScore1 = ((double)waterNorth / waterTotal) * ((double)landSouth / landTotal);
			const double divisionScore2 = ((double)waterSouth / waterTotal) * ((double)landNorth / landTotal);
			double bestDivisionScore = std::max(divisionScore1, divisionScore2);

			return (borderScore + 2.0 * bestDivisionScore) / 3.0;
		}

		// all
		// only a specific area changed:
		double computeScore()
		{
			// zunchst die Lsung auf Gltigkeit prfen
			/*for(int x = 0; x < MAP_WIDTH; ++x)
				if(border[x] <= 0 || border[x] >= MAP_HEIGHT)
				   throw std::runtime_error("Ungltiger Grenzverlauf!");*/

			unsigned int borderLength = 0;
			for(unsigned int x=0; x < MAP_WIDTH - 1; ++x)
				borderLength += absInt(border[x] - border[x + 1]);		// bit operation remove sign?
			borderLength += absInt(border[MAP_WIDTH - 1] - border[0]);
			const double borderScore = 1.0f - borderLength / 2048.0f;

			unsigned int waterNorth = 0;
			unsigned int waterSouth = 0;
			unsigned int landNorth = 0;
			unsigned int landSouth = 0;
			for(int x=0; x < MAP_WIDTH; ++x)
			{
				// north
				landNorthColumn[x] = 0;
				for(int y = 0; y < border[x]; ++y)
					landNorthColumn[x] += planetMap[x][y] ? 1 : 0;
				landNorth += landNorthColumn[x];
				unsigned int waterNorthColumn = border[x] - landNorthColumn[x];
				waterNorth += waterNorthColumn;
				// south
				waterSouth += waterPerX[x] - waterNorthColumn;
				landSouth  += landPerX[x] - landNorthColumn[x];
			}
			const double divisionScore1 = ((double)waterNorth / waterTotal) * ((double)landSouth / landTotal);
			const double divisionScore2 = ((double)waterSouth / waterTotal) * ((double)landNorth / landTotal);
			double bestDivisionScore = std::max(divisionScore1, divisionScore2);

			return (borderScore + 2.0f * bestDivisionScore) / 3.0f;
		}

		unsigned int landNorthColumn[MAP_WIDTH];
		const PlanetMap& planetMap;
		Border border;
	};

	// precalculate water/land amounts
	landTotal = 0;
	waterTotal = 0;
	for(int x = 0; x < MAP_WIDTH; ++x)
	{
		landPerX[x] = waterPerX[x] = 0;
		bool valueBefore = planetMap[x][0];
		for(int y=0; y < MAP_HEIGHT; ++y)
		{
			bool b = planetMap[x][y];
			if(b)
			{
				landTotal += 1;
				landPerX[x] += 1;
			}
			else
			{
				waterTotal += 1;
				waterPerX[x] += 1;
			}
			valueBefore = b;
		}
	}

	/////////////////////////////////////////////////////////////////////
	// SLICER
	/////////////////////////////////////////////////////////////////////
	double Score;
	{
#define SMALLEST_STEP 1
#define NUM_PARTITIONS 9

	for(int x = 0; x < MAP_WIDTH; ++x)
		outBorder[x] = MAP_HEIGHT/2;


	FastBorder tempBorder0(planetMap); memcpy(tempBorder0.border, outBorder, sizeof(int) * MAP_WIDTH);
	FastBorder tempBorder1(planetMap); memcpy(tempBorder1.border, outBorder, sizeof(int) * MAP_WIDTH);
	Score = tempBorder0.computeScore();
	int partNum = 0;

	for(int part=MAP_WIDTH; partNum < NUM_PARTITIONS; part /= 2, ++partNum)
	{
		for(int partStart=0; partStart <= MAP_WIDTH - part; partStart += part)
		{
			for(int Step = MAP_HEIGHT/2-1; Step > SMALLEST_STEP; Step /= 2)
			{
				for(int x = partStart; x < partStart+part; ++x)
				{
					tempBorder0.border[x] = (outBorder[x] + Step) % (MAP_HEIGHT-1) + 1;
					tempBorder1.border[x] = (unsigned int)(outBorder[x] - Step) % MAP_HEIGHT;
				}
				double Score0 = tempBorder0.computeScore(partStart, partStart+part);
				double Score1 = tempBorder1.computeScore(partStart, partStart+part);
				if(Score0 > Score && Score0 > Score1)
				{
					memcpy(outBorder+partStart, tempBorder0.border+partStart, sizeof(int) * part);
					memcpy(tempBorder1.border+partStart, outBorder+partStart, sizeof(int) * part);
					Score = Score0;
				}
				else if(Score1 > Score && Score1 > Score1)
				{
					memcpy(outBorder+partStart, tempBorder1.border+partStart, sizeof(int) * part);
					memcpy(tempBorder0.border+partStart, outBorder+partStart, sizeof(int) * part);
					Score = Score1;
				}
			}
		}
	}
	}

	/////////////////////////////////////////////////////////////////////
	// STEPPER - good method for maps without lakes
	/////////////////////////////////////////////////////////////////////
	FastBorder tempBorder(planetMap);
	{
#define SCALEFACTOR 32
#define MAP_WIDTH_SCALED (MAP_WIDTH / SCALEFACTOR)
#define MAP_HEIGHT_SCALED (MAP_HEIGHT / SCALEFACTOR)

	bool scaledPlanetMap[MAP_WIDTH_SCALED][MAP_HEIGHT_SCALED];
	for(int x = 0; x < MAP_WIDTH_SCALED; ++x)
	{
		int startX = x*SCALEFACTOR;
		int endX = startX + SCALEFACTOR;
		for(int y = 0; y < MAP_HEIGHT_SCALED; ++y)
		{
			int startY = y*SCALEFACTOR;
			int endY = startY + SCALEFACTOR;
			bool& mapVal = scaledPlanetMap[x][y];
			int value = 0;
			for(int _x=startX; _x<endX; ++_x)
			{
				for(int _y=startY; _y<endY; ++_y)
					value += planetMap[_x][_y] ? 1 : -1;
			}
			mapVal = value > 0.0f;
		}
	}

	// calculate water/land amounts, precalculate vertical jumps
	int landTotal = 0;
	int waterTotal = 0;
	int landPerX[MAP_WIDTH_SCALED];
	int waterPerX[MAP_WIDTH_SCALED];
	int verticalJumpingList[MAP_WIDTH_SCALED][MAP_HEIGHT_SCALED];
	memset(verticalJumpingList, 0, MAP_WIDTH_SCALED*MAP_HEIGHT_SCALED*sizeof(int));
	for(int x = 0; x < MAP_WIDTH_SCALED; ++x)
	{
		landPerX[x] = waterPerX[x] = 0;
		int yJumping = 0;
		bool valueBefore = scaledPlanetMap[x][0];
		for(int y=0; y < MAP_HEIGHT_SCALED; ++y)
		{
			bool b = scaledPlanetMap[x][y];
			if(b)
			{
				landTotal += 1;
				landPerX[x] += 1;
			}
			else
			{
				waterTotal += 1;
				waterPerX[x] += 1;
			}

			if(valueBefore == b)
				++yJumping;
			else
			{
				verticalJumpingList[x][y-yJumping] = yJumping;
				yJumping = 1;
			}

			valueBefore = b;
		}
		verticalJumpingList[x][MAP_HEIGHT_SCALED-yJumping] = yJumping-1;
	}

	double BestScoreFast = -9999999.;
	int BestBorderScaled[MAP_WIDTH_SCALED]; memset(BestBorderScaled, 0, sizeof(int) * MAP_WIDTH_SCALED);
	int TempBorderScaled[MAP_WIDTH_SCALED]; memset(TempBorderScaled, 0, sizeof(int) * MAP_WIDTH_SCALED);
	bool Continue = true;
	while(Continue)
	{
		Continue = false;
		for(int i=0; i<MAP_WIDTH_SCALED; ++i)
		{
			TempBorderScaled[i] += verticalJumpingList[i][TempBorderScaled[i]];
			if(!verticalJumpingList[i][TempBorderScaled[i]])
				TempBorderScaled[i] = MAP_HEIGHT_SCALED-1;
			else
				Continue = true;
		}

		// estimate score
		int waterNorth = 0;
		int waterSouth = 0;
		int landNorth = 0;
		int landSouth = 0;
		for(int x = 0; x < MAP_WIDTH_SCALED; ++x)
		{
			// north
			int landNorthColumn = 0;
			int waterNorthColumn = 0;
			for(int y = 0; y < TempBorderScaled[x]; ++y)
			{
				if(scaledPlanetMap[x][y])
					landNorthColumn += 1;
				else
					waterNorthColumn += 1;

			}
			waterNorth += waterNorthColumn;
			landNorth += landNorthColumn;
			// south
			waterSouth += waterPerX[x] - waterNorthColumn;
			landSouth  += landPerX[x] - landNorthColumn;
		}
		const double divisionScore1 = ((double)waterNorth / waterTotal) * ((double)landSouth / landTotal);
		const double divisionScore2 = ((double)waterSouth / waterTotal) * ((double)landNorth / landTotal);
		double bestDivisionScore = std::max(divisionScore1, divisionScore2);

		double ScoreFast = bestDivisionScore;

		// better score?
		if(ScoreFast > BestScoreFast)
		{
			BestScoreFast = ScoreFast;
			memcpy(BestBorderScaled, TempBorderScaled, sizeof(int)*MAP_WIDTH_SCALED);
		}
	}

	unsigned int x = 0;
	for(unsigned int ScaledX = 0; ScaledX<MAP_WIDTH_SCALED; ++ScaledX)
	{
		for(int i=0; i<SCALEFACTOR; ++x, ++i)
		{
			tempBorder.border[x] = BestBorderScaled[ScaledX] * SCALEFACTOR;
			tempBorder.border[x]++;
		}
	}

	double NewScore = tempBorder.computeScore();
	if(NewScore > Score)
	{
//		std::cout << "\nStepper better than slicer! Difference: " << (NewScore - Score) << std::endl;
		memcpy(outBorder, tempBorder.border, sizeof(int) * MAP_WIDTH);
		Score = NewScore;
	}
	else
	{
		memcpy(tempBorder.border, outBorder, sizeof(int) * MAP_WIDTH);
		Score = tempBorder.computeScore();
	}

	}

	/////////////////////////////////////////////////////////////////////
	// REST TIME EXPLOIT - iterative improvement
	////////////////////////////////////////////////////////////////////
	srand((unsigned int)(clock()*1000));
	// select small random areas and make a random operation
	// because of the optimized computescore function, no areas over the left/right border
#define AREA_SIZE_MAX 32
#define OPERATION_MAX 5

	unsigned int landNorthColumnBuffer[AREA_SIZE_MAX];
	while(clock() < 0.995f)
	{
		unsigned int areaStart = rand() % MAP_WIDTH;
		unsigned int areaSize = rand() % AREA_SIZE_MAX;
		unsigned int areaEnd = areaStart + areaSize;
		if(areaEnd >= MAP_WIDTH)
		{
			areaEnd = MAP_WIDTH-1;
			areaSize = areaEnd-areaStart;
		}
		int operation = rand() % (OPERATION_MAX*2) - OPERATION_MAX;

		memcpy(landNorthColumnBuffer, tempBorder.landNorthColumn+areaStart, sizeof(unsigned int) * areaSize);

		for(unsigned int i=areaStart; i<areaEnd; ++i)
			tempBorder.border[i] = (unsigned int)(tempBorder.border[i] + operation) % (MAP_HEIGHT-1) + 1;//(unsigned int)(tempBorder.border[i] + operation) % MAP_HEIGHT;

		double NewScore = tempBorder.computeScore(areaStart, areaEnd);
		if(NewScore > Score)
		{
			memcpy(outBorder+areaStart, tempBorder.border+areaStart, sizeof(int) * areaSize);
			Score = NewScore;
		}
		else
		{
			memcpy(tempBorder.border+areaStart, outBorder+areaStart, sizeof(int) * areaSize);
			memcpy(tempBorder.landNorthColumn+areaStart, landNorthColumnBuffer, sizeof(unsigned int) * areaSize);
		}
	}
}

}
